"""
footing.constants
~~~~~~~~~~~~~~~~~

Constants for footing
"""

#: The environment variable set when running any footing command. It is set to
#: the name of the command
FOOTING_ENV_VAR = '_FOOTING'

#: The footing config file in each repo
FOOTING_CONFIG_FILE = 'footing.yaml'

#: The Github API token environment variable
GITHUB_API_TOKEN_ENV_VAR = 'GITHUB_API_TOKEN'

#: The Gitlab API token environment variable
GITLAB_API_TOKEN_ENV_VAR = 'GITLAB_API_TOKEN'

#: Footing docs URL
FOOTING_DOCS_URL = 'https://github.com/CloverHealth/footing'

#: The temporary branches used for updates
UPDATE_BRANCH_NAME = '_footing_update'
TEMP_UPDATE_BRANCH_NAME = UPDATE_BRANCH_NAME + '_temp'



"""Tests for footing.update module"""
import subprocess

import pytest

import footing.constants
import footing.forge
import footing.update


def test_get_latest_template_version(mocker):
    """Tests footing.update._get_latest_template_version"""
    mocker.patch.object(
        footing.forge.Github,
        'get_latest_template_version',
        autospec=True,
        return_value='v1',
    )
    assert (
        footing.update._get_latest_template_version(
            'git@github.com:org/template.git'
        )
        == 'v1'
    )


@pytest.mark.parametrize(
    'old_config, new_config, expected_has_changed',
    [
        ('{"config": "same"}', '{"config": "same"}', False),
        ('{"config": "same"}', '{"config": "diff"}', True),
    ],
)
def test_cookiecutter_configs_have_changed(
    old_config, new_config, expected_has_changed, mocker, responses
):
    """Tests footing.update._cookiecutter_configs_have_changed"""
    mock_clone = mocker.patch(
        'footing.update.cc_vcs.clone', autospec=True, return_value='path'
    )
    mock_open = mocker.patch('footing.update.open')
    mock_open().read.side_effect = [old_config, new_config]
    mock_call = mocker.patch('subprocess.check_call', autospec=True)
    template = 'git@github.com:org/template.git'

    config_has_changed = footing.update._cookiecutter_configs_have_changed(
        template, 'old', 'new'
    )
    assert config_has_changed == expected_has_changed
    mock_clone.assert_called_once_with(template, 'old', mocker.ANY)
    assert mock_call.call_count == 1


@pytest.mark.parametrize('existing_files', [True, False])
def test_apply_template(mocker, existing_files):
    mock_td = mocker.patch('tempfile.TemporaryDirectory', autospec=True)
    mock_cc = mocker.patch(
        'cookiecutter.main.cookiecutter',
        autospec=True,
        return_value='basepath',
    )
    mock_list = mocker.patch(
        'os.listdir', autospec=True, return_value=['a', 'b']
    )
    mocker.patch('os.path.isdir', autospec=True, side_effect=[True, False])
    mocker.patch('os.path.exists', autospec=True, return_value=existing_files)
    mock_shutil_ct = mocker.patch('shutil.copytree', autospec=True)
    mock_shutil_cp = mocker.patch('shutil.copy2', autospec=True)
    mock_rmtree = mocker.patch('shutil.rmtree', autospec=True)
    mock_remove = mocker.patch('os.remove', autospec=True)

    footing.update._apply_template(
        't', '.', checkout='v1', extra_context={'c': 'tx'}
    )

    mock_cc.assert_called_once_with(
        't',
        checkout='v1',
        no_input=True,
        output_dir=mock_td.return_value.__enter__.return_value,
        extra_context={'c': 'tx'},
    )
    mock_list.assert_called_once_with(mock_cc.return_value)
    mock_shutil_ct.assert_called_once_with('basepath/a', './a')
    mock_shutil_cp.assert_called_once_with('basepath/b', './b')
    if existing_files:
        mock_rmtree.assert_called_once_with('./a')
        mock_remove.assert_called_once_with('./b')
    else:
        assert not mock_rmtree.called
        assert not mock_remove.called


@pytest.mark.parametrize(
    'footing_config, latest_version, supplied_version, expected_up_to_date',
    [
        ({'_version': 'v1', '_template': 't'}, 'v2', None, False),
        ({'_version': 'v1', '_template': 't'}, 'v1', None, True),
        ({'_version': 'v1', '_template': 't'}, 'v1', 'v2', False),
        ({'_version': 'v3', '_template': 't'}, 'v1', 'v3', True),
    ],
)
def test_up_to_date(
    footing_config,
    latest_version,
    supplied_version,
    expected_up_to_date,
    mocker,
):
    """Tests footing.update.up_to_date"""
    mocker.patch('footing.check.in_git_repo', autospec=True)
    mocker.patch('footing.check.in_clean_repo', autospec=True)
    mocker.patch('footing.check.is_footing_project', autospec=True)
    mocker.patch(
        'footing.utils.read_footing_config',
        autospec=True,
        return_value=footing_config,
    )
    mocker.patch(
        'footing.update._get_latest_template_version',
        autospec=True,
        return_value=latest_version,
    )

    assert (
        footing.update.up_to_date(version=supplied_version)
        == expected_up_to_date
    )


@pytest.mark.parametrize(
    'version, supplied_version, latest_version',
    [
        ('v1', None, 'v1'),
        ('v1', 'v1', 'v0'),
    ],
)
def test_update_w_up_to_date(
    version, supplied_version, latest_version, mocker
):
    """Tests footing.update.update when the template is already up to date"""
    mocker.patch('footing.check.not_has_branch', autospec=True)
    mocker.patch('footing.check.in_git_repo', autospec=True)
    mocker.patch('footing.check.in_clean_repo', autospec=True)
    mocker.patch('footing.check.is_footing_project', autospec=True)
    mocker.patch(
        'footing.utils.read_footing_config',
        autospec=True,
        return_value={'_version': version, '_template': 't'},
    )
    mocker.patch(
        'footing.update._get_latest_template_version',
        autospec=True,
        return_value=latest_version,
    )

    assert not footing.update.update(new_version=supplied_version)


@pytest.mark.parametrize(
    'cc_configs_changed, enter_parameters, current_version, latest_version, old_template',
    [
        (False, False, 'v1', 'v2', None),
        (False, False, 'v1', 'v2', 'git@github.com:owner/old_repo.git'),
        (True, False, 'v1', 'v2', None),
        (False, True, 'v1', 'v2', None),
        # Updates should still proceed when entering parameters and up to date with latest
        (False, True, 'v1', 'v1', None),
    ],
)
def test_update_w_out_of_date(
    cc_configs_changed,
    enter_parameters,
    current_version,
    latest_version,
    old_template,
    mocker,
    fs,
):
    """Tests footing.update.update when the template is out of date"""
    template = 'git@github.com:owner/repo.git'
    footing_config = {'_version': current_version, '_template': template}
    mocker.patch('footing.check.in_git_repo', autospec=True)
    mocker.patch('footing.check.in_clean_repo', autospec=True)
    mocker.patch('footing.check.is_footing_project', autospec=True)
    mocker.patch('footing.check.not_has_branch', autospec=True)
    mocker.patch(
        'footing.utils.read_footing_config',
        autospec=True,
        return_value=footing_config,
    )
    mocker.patch(
        'footing.update._get_latest_template_version',
        autospec=True,
        return_value=latest_version,
    )
    mock_apply_template = mocker.patch(
        'footing.update._apply_template', autospec=True
    )
    mock_cc_configs_have_changed = mocker.patch(
        'footing.update._cookiecutter_configs_have_changed',
        autospec=True,
        return_value=cc_configs_changed,
    )
    mock_input = mocker.patch('builtins.input', autospec=True)
    mock_get_cc_config = mocker.patch(
        'footing.utils.get_cookiecutter_config',
        autospec=True,
        return_value=('repo', footing_config),
    )
    mock_shell = mocker.patch('footing.utils.shell', autospec=True)
    mock_write_config = mocker.patch(
        'footing.utils.write_footing_config', autospec=True
    )

    footing.update.update(
        enter_parameters=enter_parameters, old_template=old_template
    )

    assert mock_input.called == cc_configs_changed or old_template is not None
    assert mock_get_cc_config.called == (
        cc_configs_changed or enter_parameters or old_template is not None
    )
    assert mock_apply_template.call_args_list == [
        mocker.call(
            old_template or template,
            '.',
            checkout=current_version,
            extra_context=footing_config,
        ),
        mocker.call(
            template,
            '.',
            checkout=latest_version,
            extra_context=footing_config,
        ),
    ]
    mock_write_config.assert_called_once_with(
        footing_config, template, latest_version
    )
    if not old_template:
        mock_cc_configs_have_changed.assert_called_once_with(
            template, current_version, latest_version
        )
    else:
        assert not mock_cc_configs_have_changed.called

    assert mock_shell.call_args_list == [
        mocker.call(
            'git checkout -b _footing_update', stderr=subprocess.DEVNULL
        ),
        mocker.call(
            'git checkout --orphan _footing_update_temp',
            stderr=subprocess.DEVNULL,
        ),
        mocker.call('git rm -rf .', stdout=subprocess.DEVNULL),
        mocker.call('git add .'),
        mocker.call(
            'git commit --no-verify -m "Initialize template from version {}"'.format(
                current_version
            ),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ),
        mocker.call('git checkout _footing_update', stderr=subprocess.DEVNULL),
        mocker.call(
            'git merge -s ours --no-edit --allow-unrelated-histories '
            '_footing_update_temp',
            stderr=subprocess.DEVNULL,
        ),
        mocker.call(
            'git checkout _footing_update_temp', stderr=subprocess.DEVNULL
        ),
        mocker.call('git rm -rf .', stdout=subprocess.DEVNULL),
        mocker.call('git add .'),
        mocker.call(
            'git commit --no-verify -m "Update template to version {}"'.format(
                latest_version
            ),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ),
        mocker.call('git checkout _footing_update', stderr=subprocess.DEVNULL),
        mocker.call(
            'git merge --no-commit _footing_update_temp',
            check=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ),
        mocker.call(
            'git checkout --theirs footing.yaml',
            check=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ),
        mocker.call(
            'git branch -D _footing_update_temp', stdout=subprocess.DEVNULL
        ),
    ]



from unittest import mock

import pytest

import footing.clean
import footing.exceptions


@pytest.mark.parametrize(
    'update_branch_exists, temp_update_branch_exists, expected_shell_cmds',
    [
        (False, False, []),
        (True, False, [mock.call('git branch -D _footing_update')]),
        (False, True, [mock.call('git branch -D _footing_update_temp')]),
        (
            True,
            True,
            [
                mock.call('git branch -D _footing_update'),
                mock.call('git branch -D _footing_update_temp'),
            ],
        ),
    ],
)
def test_clean(
    update_branch_exists,
    temp_update_branch_exists,
    expected_shell_cmds,
    mocker,
):
    """Tests footing.clean.clean"""

    def branch_exists_side_effect(branch_name):
        """A side effect to use when mocking out the function that checks if a branch exists"""
        if branch_name == footing.constants.UPDATE_BRANCH_NAME:
            return update_branch_exists
        elif branch_name == footing.constants.TEMP_UPDATE_BRANCH_NAME:
            return temp_update_branch_exists
        else:
            raise AssertionError

    mocker.patch(
        'footing.clean._get_current_branch', return_value='current_branch'
    )
    mocker.patch('footing.check.in_git_repo')
    mocker.patch(
        'footing.check._has_branch',
        autospec=True,
        side_effect=branch_exists_side_effect,
    )
    mock_shell = mocker.patch('footing.utils.shell', autospec=True)
    footing.clean.clean()
    assert mock_shell.call_args_list == expected_shell_cmds


def test_get_current_branch():
    """Verifies that the "_get_current_branch" function returns non-empty data

    This ensures that the proper git call is made and that output is parsed properly"""
    assert footing.clean._get_current_branch()


def test_clean_bad_current_branch(mocker):
    """Tests footing.clean.clean when on a branch that will be deleted"""
    mocker.patch('footing.check.in_git_repo')
    mocker.patch(
        'footing.clean._get_current_branch',
        return_value=footing.constants.UPDATE_BRANCH_NAME,
    )

    with pytest.raises(footing.exceptions.InvalidCurrentBranchError):
        footing.clean.clean()



"""
footing.setup
~~~~~~~~~~~~~

Creates and initializes a project from a template
"""
import subprocess
import unittest.mock

import cookiecutter.generate as cc_generate
import cookiecutter.hooks as cc_hooks

import footing.check
import footing.constants
import footing.utils


def _patched_run_hook(hook_name, project_dir, context):
    """Used to patch cookiecutter's ``run_hook`` function.

    This patched version ensures that the footing.yaml file is created before
    any cookiecutter hooks are executed
    """
    if hook_name == 'post_gen_project':
        with footing.utils.cd(project_dir):
            footing.utils.write_footing_config(
                context['cookiecutter'],
                context['template'],
                context['version'],
            )
    return cc_hooks.run_hook(hook_name, project_dir, context)


def _generate_files(repo_dir, config, template, version):
    """Uses cookiecutter to generate files for the project.

    Monkeypatches cookiecutter's "run_hook" to ensure that the footing.yaml file is
    generated before any hooks run. This is important to ensure that hooks can also
    perform any actions involving footing.yaml
    """
    with unittest.mock.patch(
        'cookiecutter.generate.run_hook', side_effect=_patched_run_hook
    ):
        cc_generate.generate_files(
            repo_dir=repo_dir,
            context={
                'cookiecutter': config,
                'template': template,
                'version': version,
            },
            overwrite_if_exists=False,
            output_dir='.',
        )


@footing.utils.set_cmd_env_var('setup')
def setup(template, version=None):
    """Sets up a new project from a template

    Note that the `footing.constants.FOOTING_ENV_VAR` is set to 'setup' during the duration
    of this function.

    Args:
        template (str): The git path to a template
        version (str, optional): The version of the template to use when updating. Defaults
            to the latest version
    """
    footing.check.not_in_git_repo()

    repo_path = footing.utils.get_repo_path(template)
    msg = (
        'You will be prompted for the parameters of your new project.'
        ' Please read the docs at https://github.com/{} before entering parameters.'
    ).format(repo_path)
    print(msg)

    cc_repo_dir, config = footing.utils.get_cookiecutter_config(
        template, version=version
    )

    if not version:
        with footing.utils.cd(cc_repo_dir):
            ret = footing.utils.shell(
                'git rev-parse HEAD', stdout=subprocess.PIPE
            )
            version = ret.stdout.decode('utf-8').strip()

    _generate_files(
        repo_dir=cc_repo_dir, config=config, template=template, version=version
    )



"""Tests for footing.cli module"""
# pylint: disable=no-value-for-parameter
import collections
import sys

import click
import pkg_resources
import pytest

import footing.cli
import footing.exceptions


@pytest.fixture
def mock_exit(mocker):
    yield mocker.patch('sys.exit', autospec=True)


@pytest.fixture
def mock_successful_exit(mock_exit):
    yield
    mock_exit.assert_called_once_with(0)


@pytest.mark.usefixtures('mock_successful_exit')
@pytest.mark.parametrize(
    'subcommand, args, expected_function, exp_call_args, exp_call_kwargs',
    [
        ('setup', ['t'], 'footing.setup.setup', ['t'], {'version': None}),
        (
            'setup',
            ['t', '-v', 'v1'],
            'footing.setup.setup',
            ['t'],
            {'version': 'v1'},
        ),
        (
            'update',
            [],
            'footing.update.update',
            [],
            {'new_version': None, 'enter_parameters': False},
        ),
        (
            'update',
            ['-c', '-v', 'v1'],
            'footing.update.up_to_date',
            [],
            {'version': 'v1'},
        ),
        ('ls', ['user'], 'footing.ls.ls', ['user'], {'template': None}),
        (
            'ls',
            ['user', 'template'],
            'footing.ls.ls',
            ['user'],
            {'template': 'template'},
        ),
        ('clean', [], 'footing.clean.clean', [], {}),
        (
            'switch',
            ['new_t', '-v', 'new_v'],
            'footing.update.update',
            [],
            {'new_template': 'new_t', 'new_version': 'new_v'},
        ),
    ],
)
def test_main(
    subcommand, args, expected_function, exp_call_args, exp_call_kwargs, mocker
):
    """Verify calling the CLI subcommands works as expected"""
    mocker.patch.object(sys, 'argv', ['footing', subcommand] + args)
    mock_expected_func = mocker.patch(expected_function, autospec=True)

    footing.cli.main()

    mock_expected_func.assert_called_once_with(
        *exp_call_args, **exp_call_kwargs
    )


@pytest.mark.usefixtures('mock_successful_exit')
def test_main_w_version(mocker, capsys):
    """Test calling the CLI with the --version option"""
    mocker.patch.object(sys, 'argv', ['footing', '--version'])

    footing.cli.main()

    out, _ = capsys.readouterr()
    version = pkg_resources.get_distribution('footing').version
    assert out == 'footing %s\n' % version


@pytest.mark.usefixtures('mock_successful_exit')
def test_main_no_args(mocker, capsys):
    """Test calling the CLI with no options"""
    mocker.patch.object(sys, 'argv', ['footing'])
    mocker.patch.object(
        click.Context, 'get_help', autospec=True, return_value='help_text'
    )

    footing.cli.main()

    out, _ = capsys.readouterr()
    assert out == 'help_text\n'


@pytest.mark.usefixtures('mock_exit')
@pytest.mark.parametrize(
    'version, up_to_date_return',
    [
        pytest.param(
            None,
            False,
            marks=pytest.mark.xfail(
                raises=footing.exceptions.NotUpToDateWithTemplateError
            ),
        ),
        (None, True),
        ('version', True),
    ],
)
def test_update_check(version, up_to_date_return, capsys, mocker):
    """Verifies checking for updates when calling footing update -c"""
    mocker.patch.object(
        sys, 'argv', ['footing', 'update', '-c', '-v', version]
    )
    mock_up_to_date = mocker.patch(
        'footing.update.up_to_date',
        autospec=True,
        return_value=up_to_date_return,
    )

    footing.cli.main()

    out, _ = capsys.readouterr()
    assert out == 'Footing package is up to date\n'
    mock_up_to_date.assert_called_once_with(version=version)


@pytest.mark.usefixtures('mock_successful_exit')
@pytest.mark.parametrize(
    'ls_args, expected_out',
    [
        (['footing', 'ls', 'user'], 'ls\nvalues\n'),
        (
            ['footing', 'ls', 'user', '-l'],
            'ls - ls descr\nvalues - (no project description found)\n',
        ),
    ],
)
def test_ls(ls_args, expected_out, capsys, mocker):
    """Verify ls prints results properly"""
    mocker.patch(
        'footing.ls.ls',
        autospec=True,
        return_value=collections.OrderedDict(
            [
                ('ls', 'ls descr'),
                ('values', '(no project description found)'),
            ]
        ),
    )
    mocker.patch.object(sys, 'argv', ls_args)

    footing.cli.main()

    out, _ = capsys.readouterr()
    assert out == expected_out



"""Tests for footing.setup module"""
import os
import subprocess

import pytest

import footing.constants
import footing.exceptions
import footing.setup


@pytest.mark.parametrize(
    'version, expected_revparse_called, expected_version',
    [
        (None, True, 'latest_version'),
        ('version', False, 'version'),
    ],
)
def test_setup(version, expected_revparse_called, expected_version, mocker):
    """Tests footing.setup.setup_template"""
    config = {'my': 'config'}
    template = 'git@github.com:user/template.git'
    mocker.patch('footing.check.not_in_git_repo', autospec=True)
    mock_get_cc_config = mocker.patch(
        'footing.utils.get_cookiecutter_config',
        autospec=True,
        return_value=('.', config),
    )

    revparse_return = subprocess.CompletedProcess(
        [], stdout=b'latest_version', returncode=0
    )
    mock_generate_files = mocker.patch(
        'footing.setup.cc_generate.generate_files',
        autospec=True,
        return_value='.',
    )
    mock_shell = mocker.patch(
        'footing.utils.shell',
        autospec=True,
        side_effect=[revparse_return, None, None, None, None, None],
    )

    footing.setup.setup(template, version=version)

    mock_get_cc_config.assert_called_once_with(template, version=version)
    mock_generate_files.assert_called_once_with(
        context={
            'cookiecutter': config,
            'template': template,
            'version': expected_version,
        },
        output_dir='.',
        overwrite_if_exists=False,
        repo_dir='.',
    )
    assert mock_shell.called == expected_revparse_called


def test_generate_files(tmpdir):
    """Generate files for a fake template created in a temporary directory"""
    template_dir = '%s/template' % tmpdir
    template_project_dir = '%s/{{cookiecutter.name}}' % template_dir
    hooks_dir = '%s/template/hooks' % tmpdir
    project_dir = '%s/project' % tmpdir

    os.mkdir(template_dir)
    os.mkdir(template_project_dir)
    os.mkdir(hooks_dir)
    with open('%s/cookiecutter.json' % template_dir, 'w') as config:
        config.write('{"name": null}')
    with open('%s/name.txt' % template_project_dir, 'w') as name:
        name.write('my name is {{cookiecutter.name}}')
    with open('%s/post_gen_project.sh' % hooks_dir, 'w') as hook:
        hook.write('#!/usr/bin/env sh\ntouch hook_file\n')
    os.chmod('%s/post_gen_project.sh' % hooks_dir, 0o700)

    with footing.utils.cd(str(tmpdir)):
        footing.setup._generate_files(
            repo_dir=template_dir,
            config={'name': 'project'},
            template='template',
            version='version',
        )

    # Verify files were created properly
    with footing.utils.cd(project_dir):
        with open('name.txt') as name:
            assert name.read() == 'my name is project'
        config = footing.utils.read_footing_config()
        assert config == {
            '_template': 'template',
            '_version': 'version',
            'name': 'project',
        }
        # The post_gen_project hook should have made this file
        assert os.path.exists('hook_file')



import collections

import pytest

import footing.forge
import footing.ls


@pytest.mark.parametrize(
    'forge, template, github_query',
    [
        ('github.com/u', None, 'user:u filename:cookiecutter.json'),
        ('github.com/u/', None, 'user:u filename:cookiecutter.json'),
        (
            'github.com/u',
            'git@github.com:u/t.git',
            'user:u filename:footing.yaml t',
        ),
    ],
)
def test_ls(forge, template, github_query, mocker):
    mock_code_search = mocker.patch.object(
        footing.forge.Github,
        '_code_search',
        autospec=True,
        return_value={
            'repo2': {'description': 'description 2'},
            'repo1': {'description': 'description 1'},
        },
    )

    results = footing.ls.ls(forge, template=template)

    mock_code_search.assert_called_once_with(
        mocker.ANY, github_query, forge=forge
    )
    assert results == collections.OrderedDict(
        [
            ('repo1', 'description 1'),
            ('repo2', 'description 2'),
        ]
    )



"""Tests fr footing.utils module"""
import os
import subprocess

import pytest

import footing.constants
import footing.utils


@pytest.mark.parametrize(
    'cmd, check, stdin, stdout, stderr',
    [
        ('cmd', True, subprocess.PIPE, subprocess.PIPE, subprocess.PIPE),
        ('cmd', False, None, None, None),
    ],
)
def test_shell(cmd, check, stdin, stdout, stderr, mocker):
    """Tests footing.utils.shell"""
    mock_run = mocker.patch('subprocess.run', autospec=True)

    footing.utils.shell(
        cmd, check=check, stdin=stdin, stdout=stdout, stderr=stderr
    )

    mock_run.assert_called_once_with(
        cmd, shell=True, check=check, stdin=stdin, stdout=stdout, stderr=stderr
    )


def test_cd(fs):
    """Tests footing.utils.cd using a fake filesystem"""
    os.mkdir('/tmp_dir')
    orig_cwd = os.getcwd()
    with footing.utils.cd('/tmp_dir'):
        assert os.getcwd() == '/tmp_dir'

    assert os.getcwd() == orig_cwd


def test_read_footing_config(fs):
    """Tests footing.utils.read_footing_config with a fake file system"""
    footing_config_yaml = (
        '_version: version\n'
        'repo_name: repo_name\n'
        '_extensions: [jinja2_time.TimeExtension]\n'
    )
    fs.CreateFile('footing.yaml', contents=footing_config_yaml)

    assert footing.utils.read_footing_config() == {
        '_version': 'version',
        'repo_name': 'repo_name',
        '_extensions': ['jinja2_time.TimeExtension'],
    }


def test_write_footing_config(fs):
    """Tests footing.utils.write_footing_config with a fake file system"""
    footing.utils.write_footing_config(
        {
            'repo_name': 'repo_name',
            '_extensions': ['jinja2_time.TimeExtension'],
        },
        template='t',
        version='version',
    )

    with open(footing.constants.FOOTING_CONFIG_FILE) as config:
        assert set(config.readlines()) == {
            '_extensions:\n',
            '- jinja2_time.TimeExtension\n',
            '_template: t\n',
            '_version: version\n',
            'repo_name: repo_name\n',
        }


@pytest.mark.parametrize(
    'default_config',
    [
        None,
        {'my': 'config'},
    ],
)
def test_get_cookiecutter_config(default_config, mocker):
    """Tests footing.utils.get_cookiecutter_config"""
    default_context = {'default': 'context'}
    generated_context = {'generated': 'context'}
    prompted_context = {'prompted': 'context'}

    mock_get_user_conf = mocker.patch(
        'cookiecutter.config.get_user_config',
        autospec=True,
        return_value={
            'abbreviations': 'abbr',
            'cookiecutters_dir': 'cc_dir',
            'default_context': default_context,
        },
    )
    mock_repo_dir = mocker.patch(
        'cookiecutter.repository.determine_repo_dir',
        autospec=True,
        return_value=('repo_dir', True),
    )
    mock_gen_context = mocker.patch(
        'cookiecutter.generate.generate_context',
        autospec=True,
        return_value=generated_context,
    )

    mock_prompt_for_conf = mocker.patch(
        'cookiecutter.prompt.prompt_for_config',
        autospec=True,
        return_value=prompted_context,
    )

    assert footing.utils.get_cookiecutter_config(
        't', default_config=default_config
    ) == (  # noqa
        'repo_dir',
        prompted_context,
    )
    mock_get_user_conf.assert_called_once_with()
    mock_repo_dir.assert_called_once_with(
        template='t',
        abbreviations='abbr',
        clone_to_dir='cc_dir',
        checkout=None,
        no_input=True,
    )
    mock_gen_context.assert_called_once_with(
        context_file='repo_dir/cookiecutter.json',
        default_context={
            **(default_config or {}),
            **default_context,
        },
    )
    mock_prompt_for_conf.assert_called_once_with(generated_context)


def test_set_cmd_env_var():
    """Tests footing.utils.set_cmd_env_var"""

    @footing.utils.set_cmd_env_var("value1")
    def func1():
        assert os.environ[footing.constants.FOOTING_ENV_VAR] == 'value1'
        return 123

    @footing.utils.set_cmd_env_var("value2")
    def func2():
        assert os.environ[footing.constants.FOOTING_ENV_VAR] == 'value2'
        raise NotImplementedError()

    assert footing.constants.FOOTING_ENV_VAR not in os.environ
    assert func1() == 123
    assert footing.constants.FOOTING_ENV_VAR not in os.environ
    with pytest.raises(NotImplementedError):
        func2()
    assert footing.constants.FOOTING_ENV_VAR not in os.environ

    os.environ[footing.constants.FOOTING_ENV_VAR] = 'testvalue'
    try:
        assert func1() == 123
        assert os.environ[footing.constants.FOOTING_ENV_VAR] == 'testvalue'
        with pytest.raises(NotImplementedError):
            func2()
        assert os.environ[footing.constants.FOOTING_ENV_VAR] == 'testvalue'
    finally:
        os.environ.pop(footing.constants.FOOTING_ENV_VAR, None)



"""Tests for footing.ls module"""
import os
import subprocess
import urllib

import pytest
import requests
from requests import codes as http_codes

import footing.constants
import footing.exceptions
import footing.forge


@pytest.mark.parametrize(
    "gitlab_api_token",
    [
        "token",
        pytest.param(
            "",
            marks=pytest.mark.xfail(
                raises=footing.exceptions.InvalidEnvironmentError
            ),
        ),
    ],
)
def test_gitlab_get_client(gitlab_api_token, mocker):
    """Tests footing.forget.Gitlab.get_client()"""
    mocker.patch.dict(
        os.environ, {"GITLAB_API_TOKEN": gitlab_api_token}, clear=True
    )

    assert footing.forge.Gitlab().get_client("https://gitlab.com")


def test_gitlab_get_gitlab_url_and_repo_path():
    """Tests footing.forge.Gitlab._get_gitlab_url_and_repo_path"""
    assert footing.forge.Gitlab()._get_gitlab_url_and_repo_path(
        "git@gitlab.com:group/path.git"
    ) == ("https://gitlab.com", "group/path")


@pytest.mark.parametrize(
    "forge, expected_results",
    [
        ("gitlab.com/my/group", ("https://gitlab.com", "my/group")),
        ("gitlab.com/my/group/", ("https://gitlab.com", "my/group")),
        ("http://gitlab.com/my/group", ("http://gitlab.com", "my/group")),
        pytest.param(
            "gitlab.com",
            None,
            marks=pytest.mark.xfail(
                raises=footing.exceptions.InvalidGitlabGroupError
            ),
        ),
    ],
)
def test_gitlab_get_gitlab_url_and_group(forge, expected_results):
    """Tests footing.forge.Gitlab._get_gitlab_url_and_group"""
    assert (
        footing.forge.Gitlab()._get_gitlab_url_and_group(forge)
        == expected_results
    )


@pytest.mark.parametrize(
    "http_status",
    [
        http_codes.ok,
        pytest.param(
            http_codes.not_found,
            marks=pytest.mark.xfail(raises=requests.exceptions.HTTPError),
        ),
    ],
)
def test_github_get_latest_template_version_api(
    http_status, mocker, responses
):
    """Tests footing.forge.Github._get_latest_template_version"""
    api = "https://api.github.com/repos/owner/template/commits"
    responses.add(responses.GET, api, json=[{"sha": "v1"}], status=http_status)

    latest = footing.forge.Github()._get_latest_template_version(
        "git@github.com:owner/template.git"
    )
    assert latest == "v1"


@pytest.mark.parametrize(
    "stdout, stderr, expected",
    [
        (b"version\n", b"", "version"),
        (b"version\n", b"stderr_can_be_there_w_stdout", "version"),
        pytest.param(
            b"\n",
            b"stderr_w_no_stdout_is_an_error",
            None,
            marks=pytest.mark.xfail(raises=RuntimeError),
        ),
    ],
)
def test_get_latest_template_version_w_ssh(mocker, stdout, stderr, expected):
    """Tests footing.forge._get_latest_template_version_w_ssh"""
    ls_remote_return = subprocess.CompletedProcess(
        [], returncode=0, stdout=stdout, stderr=stderr
    )
    mock_shell = mocker.patch(
        "footing.utils.shell", autospec=True, return_value=ls_remote_return
    )

    assert footing.forge._get_latest_template_version_w_ssh("t") == expected
    cmd = "git ls-remote t | grep HEAD | cut -f1"
    mock_shell.assert_called_once_with(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )


@pytest.mark.parametrize(
    "git_ssh_side_effect, git_api_side_effect, expected",
    [
        (["version"], None, "version"),
        (
            subprocess.CalledProcessError(returncode=1, cmd="cmd"),
            ["version"],
            "version",
        ),
        pytest.param(
            subprocess.CalledProcessError(returncode=1, cmd="cmd"),
            requests.exceptions.RequestException,
            None,
            marks=pytest.mark.xfail(raises=footing.exceptions.CheckRunError),
        ),
        pytest.param(
            subprocess.CalledProcessError(returncode=1, cmd="cmd"),
            footing.exceptions.InvalidEnvironmentError,
            None,
            marks=pytest.mark.xfail(raises=footing.exceptions.CheckRunError),
        ),
    ],
)
def test_github_get_latest_template_version(
    mocker, git_ssh_side_effect, git_api_side_effect, expected
):
    mocker.patch(
        "footing.forge._get_latest_template_version_w_ssh",
        autospec=True,
        side_effect=git_ssh_side_effect,
    )
    mocker.patch.object(
        footing.forge.Github,
        "_get_latest_template_version",
        autospec=True,
        side_effect=git_api_side_effect,
    )

    version = footing.forge.Github().get_latest_template_version("template")
    assert version == expected


@pytest.mark.parametrize(
    "root, expected_client_cls",
    [
        ("github.com/user", footing.forge.Github),
        ("gitlab.com/user", footing.forge.Gitlab),
        pytest.param(
            "invalid",
            None,
            marks=pytest.mark.xfail(
                raises=footing.exceptions.InvalidForgeError
            ),
        ),
    ],
)
def test_from_path(root, expected_client_cls):
    client = footing.forge.from_path(root)
    assert client.__class__ == expected_client_cls


@pytest.mark.parametrize(
    "headers, expected_links",
    [
        ({"no_link_keys": "value"}, {}),
        (
            {
                "link": '<https://url.com>; rel="next", <https://url2.com>; rel="last"'
            },
            {"next": "https://url.com", "last": "https://url2.com"},
        ),
    ],
)
def test_github_parse_link_header(headers, expected_links):
    """Tests footing.forge.Github._parse_link_header"""
    assert footing.forge.Github()._parse_link_header(headers) == expected_links


def test_github_code_search_single_page(mocker, responses):
    """Tests ls.ls for a single page of responses"""
    response_content = {
        "items": [
            {
                "repository": {
                    "full_name": "repo/repo1",
                },
            },
            {
                "repository": {
                    "full_name": "repo/repo2",
                },
            },
        ],
    }
    responses.add(
        responses.GET,
        "https://api.github.com/search/code",
        status=requests.codes.ok,
        json=response_content,
    )

    query = 'user:user {} in:path "template_path" in:file'.format(
        footing.constants.FOOTING_CONFIG_FILE
    )
    repos = footing.forge.Github()._code_search(query)

    assert repos == {
        "git@github.com:repo/repo1.git": {"full_name": "repo/repo1"},
        "git@github.com:repo/repo2.git": {"full_name": "repo/repo2"},
    }
    assert len(responses.calls) == 1
    url = urllib.parse.urlparse(responses.calls[0].request.url)
    parsed_query = urllib.parse.parse_qs(url.query)
    assert parsed_query == {
        "per_page": ["100"],
        "q": [query],
    }


def test_github_code_search_multiple_pages(mocker, responses):
    """Tests footing.forge.Github._code_search for a single page of responses"""
    response_content1 = {
        "items": [
            {
                "repository": {
                    "full_name": "repo/repo1",
                },
            },
            {
                "repository": {
                    "full_name": "repo/repo2",
                },
            },
        ],
    }
    response_link_header1 = {
        "link": '<https://next_url.com>; rel="next", <https://next_url2.com>; rel="last"',
    }
    response_content2 = {
        "items": [
            {
                "repository": {
                    "full_name": "repo/repo3",
                },
            },
            {
                "repository": {
                    "full_name": "repo/repo4",
                },
            },
        ],
    }
    responses.add(
        responses.GET,
        "https://api.github.com/search/code",
        status=requests.codes.ok,
        json=response_content1,
        adding_headers=response_link_header1,
    )
    responses.add(
        responses.GET,
        "https://next_url.com",
        status=requests.codes.ok,
        json=response_content2,
    )

    repos = footing.forge.Github()._code_search("query")

    assert repos == {
        "git@github.com:repo/repo1.git": {"full_name": "repo/repo1"},
        "git@github.com:repo/repo2.git": {"full_name": "repo/repo2"},
        "git@github.com:repo/repo3.git": {"full_name": "repo/repo3"},
        "git@github.com:repo/repo4.git": {"full_name": "repo/repo4"},
    }


@pytest.mark.parametrize(
    "ssh_path, expected_name",
    [
        ("git@github.com:user/template.git", "template"),
        ("git@github.com:user/foo-bar.git", "foo-bar"),
        ("git@github.com:user/foo_bar1.git", "foo_bar1"),
    ],
)
def test_get_name_from_ssh_path(ssh_path, expected_name):
    assert footing.forge.get_name_from_ssh_path(ssh_path) == expected_name


def test_github_code_search_invalid_root(mocker, responses):
    """Tests footing.forge.Github._code_search when the root is invalid"""
    responses.add(
        responses.GET,
        "https://api.github.com/search/code",
        status=requests.codes.unprocessable_entity,
        json={},
    )

    query = 'user:user {} in:path "template_path" in:file'.format(
        footing.constants.FOOTING_CONFIG_FILE
    )
    with pytest.raises(footing.exceptions.InvalidForgeError):
        footing.forge.Github()._code_search("invalid", query)



"""Footing test setup and fixtures"""
import os

import pytest
import responses as responses_lib

import footing.constants


@pytest.fixture
def github_env(mocker):
    """Sets Github environment variables"""
    gh_env = {footing.constants.GITHUB_API_TOKEN_ENV_VAR: 'test_gh_token'}
    mocker.patch.dict(os.environ, gh_env)
    return gh_env


@pytest.fixture(autouse=True)
def footing_env(github_env):
    """Provides a complete test footing environment for testing"""
    return github_env


@pytest.fixture
def responses():
    """Ensure no http requests happen and allow for mocking out responses"""
    with responses_lib.RequestsMock(
        assert_all_requests_are_fired=False
    ) as mocked_requests:
        yield mocked_requests



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['footing', 'footing.tests']

package_data = \
{'': ['*']}

install_requires = \
['click>=6.7',
 'cookiecutter>=1.6.0',
 'python-gitlab>=2.10.1',
 'pyyaml>=3.12',
 'requests>=2.13.0',
 'tldextract>=3.1.2']

entry_points = \
{'console_scripts': ['footing = footing.cli:main']}

setup_kwargs = {
    'name': 'footing',
    'version': '0.1.0',
    'description': 'Keep templated projects in sync with their template',
    'long_description': 'footing\n########################################################################\n\nDocumentation\n=============\n\n`View the footing docs here\n<https://footing.readthedocs.io/>`_.\n\nInstallation\n============\n\nInstall footing with::\n\n    pip3 install footing\n\n\nContributing Guide\n==================\n\nFor information on setting up footing for development and\ncontributing changes, view `CONTRIBUTING.rst <CONTRIBUTING.rst>`_.\n',
    'author': 'Opus 10 Engineering',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Opus10/footing',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4',
}


setup(**setup_kwargs)



"""Tests for footing.check module"""
import os
import subprocess

import pytest

import footing.check
import footing.constants
import footing.exceptions


def test_is_git_ssh_path_valid():
    footing.check.is_git_ssh_path('git@github.com:user/template.git')


@pytest.mark.parametrize(
    'invalid_template_path',
    [
        'bad_path',
        'git@github.com:user/template',
        '',
        'abc@def.com:user/template.git',
    ],
)
def test_is_git_ssh_path_invalid(invalid_template_path):
    with pytest.raises(footing.exceptions.InvalidTemplatePathError):
        footing.check.is_git_ssh_path(invalid_template_path)


@pytest.mark.parametrize(
    'revparse_returncode',
    [
        pytest.param(
            255,
            marks=pytest.mark.xfail(
                raises=footing.exceptions.NotInGitRepoError
            ),
        ),
        0,
    ],
)
def test_in_git_repo(revparse_returncode, mocker):
    """Tests footing.check.not_in_git_repo"""
    revparse_return = subprocess.CompletedProcess(
        [], returncode=revparse_returncode
    )
    mock_shell = mocker.patch(
        'footing.utils.shell', autospec=True, return_value=revparse_return
    )

    assert footing.check.in_git_repo() is None

    mock_shell.assert_called_once_with(
        'git rev-parse', stderr=subprocess.DEVNULL, check=False
    )


@pytest.mark.parametrize(
    'revparse_returncode',
    [
        255,
        pytest.param(
            0,
            marks=pytest.mark.xfail(raises=footing.exceptions.InGitRepoError),
        ),
    ],
)
def test_not_in_git_repo(revparse_returncode, mocker):
    """Tests footing.check.not_in_git_repo"""
    revparse_return = subprocess.CompletedProcess(
        [], returncode=revparse_returncode
    )
    mock_shell = mocker.patch(
        'footing.utils.shell', autospec=True, return_value=revparse_return
    )

    assert footing.check.not_in_git_repo() is None

    mock_shell.assert_called_once_with(
        'git rev-parse', stderr=subprocess.DEVNULL, check=False
    )


@pytest.mark.parametrize(
    'revparse_returncode',
    [
        0,
        pytest.param(
            255,
            marks=pytest.mark.xfail(
                raises=footing.exceptions.InDirtyRepoError
            ),
        ),
    ],
)
def test_in_clean_repo(revparse_returncode, mocker):
    """Tests footing.check.in_clean_repo"""
    revparse_return = subprocess.CompletedProcess(
        [], returncode=revparse_returncode
    )
    mock_shell = mocker.patch(
        'footing.utils.shell', autospec=True, return_value=revparse_return
    )

    assert footing.check.in_clean_repo() is None

    mock_shell.assert_called_once_with(
        'git diff-index --quiet HEAD --', check=False
    )


@pytest.mark.parametrize(
    'revparse_returncode',
    [
        128,
        pytest.param(
            0,
            marks=pytest.mark.xfail(
                raises=footing.exceptions.ExistingBranchError
            ),
        ),
    ],
)
def test_not_has_branch(revparse_returncode, mocker):
    """Tests footing.check.not_has_branch"""
    revparse_return = subprocess.CompletedProcess(
        [], returncode=revparse_returncode
    )
    mock_shell = mocker.patch(
        'footing.utils.shell', autospec=True, return_value=revparse_return
    )

    assert footing.check.not_has_branch('somebranch') is None

    mock_shell.assert_called_once_with(
        'git rev-parse --verify somebranch',
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        check=False,
    )


@pytest.mark.parametrize(
    'envvar_names, check_envvar_names',
    [
        pytest.param(
            ['v1', 'v2'],
            ['v2', 'v3'],
            marks=pytest.mark.xfail(
                raises=footing.exceptions.InvalidEnvironmentError
            ),
        ),
        pytest.param(
            [],
            ['v2'],
            marks=pytest.mark.xfail(
                raises=footing.exceptions.InvalidEnvironmentError
            ),
        ),
        (['v2'], ['v2']),
        (['v1', 'v2'], ['v1', 'v2']),
    ],
)
def test_has_env_vars(envvar_names, check_envvar_names, mocker):
    """Tests footing.check.has_circleci_api_token"""
    mocker.patch.dict(
        os.environ,
        {var_name: 'value' for var_name in envvar_names},
        clear=True,
    )

    assert footing.check.has_env_vars(*check_envvar_names) is None


@pytest.mark.parametrize(
    'footing_file',
    [
        pytest.param(
            'regular_file',
            marks=pytest.mark.xfail(
                raises=footing.exceptions.InvalidFootingProjectError
            ),
        ),
        footing.constants.FOOTING_CONFIG_FILE,
    ],
)
def test_check_is_footing_project(footing_file, fs):
    """Tests update._check_is_footing_project"""
    fs.CreateFile(footing_file)

    assert footing.check.is_footing_project() is None
